unit uWebshots;

interface

uses
  Windows,
  Messages,
  SysUtils,
//  Graphics, //can't use Graphics -- TC dies if we do
  ActiveX,
  Classes;

const
  wsftUnknown = 0;
  wsftWBC     = 1;
  wsftWBZ     = 2;
  wsftSPV     = 3;
  wsftWBD     = 4;
  wsftWB1     = 5;

(*

WBC file (
  WBCHeader;
  Zero padding up to file position WBCHeader.FirstPic;

  for i := 1 to WBCHeader.FileCount (
    WBCPicHeader;
    Picture data; (WBCPicHeader.PicSize)
    Thumbnail; (WBCPicHeader.ThumbnailSize)
    if i < WBCHeader.FileCount then
      Zero padding up to file position WBCHeader.DirEntry[i+1];
  )
)

*)

type
  TWBCDirEntry = packed record
    Start: Cardinal;
    Length: Cardinal;
    _unknown: Cardinal;
    Date: Cardinal; //time_t, Unix date
    Copyrighted: Cardinal;
    _unused: array[0..4] of Cardinal;
  end;

  TWBCPicHeader = packed record
    ID: Cardinal; // = $F071CDE2
    HeaderSize: Cardinal; // Size of this header
    Size: Cardinal;
// Size of the whole thing incl. this header,
// i.e. for the next header, skip to
// (Pos of Current WBCPicHeader) + Size
    FileName: array [0..255] of AnsiChar; // original file name
    Title:  array [0..127] of AnsiChar; // title of pic
    Description: array [0..255] of AnsiChar; // short description
    Credit: array [0..255] of AnsiChar; // author, etc.
    Kind: array [0..7] of AnsiChar; // 'jpg', etc.
    PicSize: Cardinal;
    ThumbnailSize: Cardinal; // Normally $00002B46 = 11078
    unk2: Cardinal; //$00000003
    unk3: array[1..140] of Byte; //00
    DateAdded: Cardinal; //time_t, Unix date
    FitToScreen: Cardinal; //0 - Use settings default; 1 - Never; 2 - Always
    ImagePID: array [0..1011] of AnsiChar; //Image ID; 'W' + image no. for Webshots images
  end;

  TWBCHeader = packed record
    ID: Cardinal; // = $95FA16AB
    FirstPic: Cardinal; // File position of first picture = Header size
    _unknown: Cardinal; // Version? Count of something? Usually 1
    Title: array[0..2183] of AnsiChar; // Catalog title
    FileCount: Cardinal; // Number of picture files in catalog
    DirEntry: array[1..160] of TWBCDirEntry;
      // array size seems to be fixed,
      // but may be variable... ( -> FileCount, FirstPic)
    _unknown2: array[0..1] of Cardinal; //zero padding?
  end;

  TWBZPicHeader = packed record
    ID: Cardinal; // = $1082CDE1
    HeaderSize: Cardinal; // Size of this header
    _unknown: Cardinal; // 1 for all wbz's I've seen... Copyright?
    FileName: array [0..255] of AnsiChar; // original file name
    DataSize: Cardinal; //Size of picture in bytes
    _unknown2: Cardinal; // 0
    DataSize2: Cardinal; //? = PicSize
    _unknown3: array [0..255] of AnsiChar; // 0
  end;

  TWBZHeader = packed record
    ID: Cardinal; // = $6791AB43
    FirstPic: Cardinal; // File position of first picture = Header size
    _unknown: Cardinal; // ?
    FileCount: Cardinal; // Number of picture files in catalog?
    EntrySize: Cardinal; //Size of picture entry (including WBZPicHeader and WBX description)
    _unknown2: Cardinal; // 2
    _unknown3: array[1..32] of Cardinal; // 0
  end;

  TWSEntry = record
    Date: TDateTime;
    FileName: array [0..255] of AnsiChar; // original file name
    Title: array [0..127] of AnsiChar; // title of pic
    Description: array [0..255] of AnsiChar; // short description
    Credit: array [0..255] of AnsiChar; // author, etc.
    Kind: array [0..7] of AnsiChar; // 'jpg', etc.
    DataSize: Cardinal;
  end;

  TWebShots = class(TObject)
  private
    FFileName: TFileName;
    FStream: TStream;
    FID: Cardinal;
    FIDStr: array[0..7] of AnsiChar;
    FFileType: Cardinal;
    FCount: Integer;
    FWBCHeader: TWBCHeader; //let's keep this in memory
    FWBZHeader: TWBZHeader; //let's keep this in memory

(*    function GetID:Cardinal;
    function GetIDStr:String;*)
    function GetFileType:Cardinal;
    procedure SetFileType(FileType: Cardinal);
    function GetHandle: THandle;
    procedure EncodeHeader(var instream,outstream: TStream);
    procedure CreateEmpty;
  public
    constructor Create(const FileName: TFileName; FileType: Cardinal = wsftUnknown);
    destructor Destroy; override;

    function GetEntry(PicIndex: Integer; var Entry: TWSEntry): Boolean;
    function Extract(PicIndex: Integer; FileName: TFileName): Boolean;
    function Add(FileName: TFileName): Boolean;
  published
    property Handle: THandle read GetHandle;
    property FileName: TFileName read FFileName;
    property FileType: Cardinal read GetFileType write SetFileType;
    property Count: Integer read FCount;
  end;

implementation

const
  idWBC    = $95FA16AB;
  idWBCPic = $F071CDE2;

  idWBZ    = $6791AB43;
  idWBZPic = $1082CDE1;

  idSPV    = 'SSPPVV00';
  idWBD    = 'WWBB0000';
  idWB1    = 'WWBB1111';


function DateTimeToUnix(const AValue: TDateTime): Int64;
begin
  Result := Round((AValue - UnixDateDelta) * SecsPerDay);
end;

function UnixToDateTime(const AValue: Int64): TDateTime;
begin
  Result := AValue / SecsPerDay + UnixDateDelta;
end;


function SaveBitmapToStream(hBitmap: HBitmap; Stream: TStream): Longint;
var
  stPictDesc: PICTDESC;
  Picture: IPicture;
begin
  Result := 0;

  stPictDesc.cbSizeofstruct := SizeOf(PICTDESC);
  stPictDesc.picType := PICTYPE_BITMAP;
  stPictDesc.hbitmap := hBitmap;
  stPictDesc.hpal := 0;

  if OleCreatePictureIndirect(stPictDesc, IPicture, FALSE, Picture) = S_OK then
    Picture.SaveAsFile(TStreamAdapter.Create(Stream),TRUE,Result);
  // clean up
//  Picture._Release;
end;

function SaveThumbnailToStream(InStream, OutStream: TStream): Boolean;
var
  FPicture: IPicture;
  r: HRESULT;
  compDC, tempDC: HDC;
  hmWidth: OLE_XSIZE_HIMETRIC;
  hmHeight: OLE_YSIZE_HIMETRIC;
  dwWidth, dwHeight: Longint;
  tempBitmap, OldBitmap: HBITMAP;
begin
  Result := False;
  if OleLoadPicture(TStreamAdapter.Create(InStream), 0, True, IPicture, FPicture) = S_OK then begin
    // get a DC to work with
    compDC := GetDC(0); // screen DC
    tempDC := CreateCompatibleDC(compDC);
    if tempDC <> 0 then begin
      // read out the width and height of the IPicture object
      FPicture.get_Width(hmWidth);
      FPicture.get_Height(hmHeight);
      dwWidth := 70;  dwHeight := 52; // standard WBC thumbnail size
      tempBitmap := CreateCompatibleBitmap(compDC, dwWidth, dwHeight);
      if tempBitmap = 0 then
        DeleteDC(tempDC)
      else begin
        OldBitmap := SelectObject(tempDC, tempBitmap);
        if OldBitmap = 0 then begin // whoops, didn't select our bitmap
          DeleteDC(tempDC);
          DeleteObject(tempBitmap);
        end
        else begin
          // ok, now we have our bitmap mounted onto our temporary DC, let's blit to it
          r := FPicture.Render(tempDC, 0, 0, dwWidth, dwHeight, 0, hmHeight-1, hmWidth, -hmHeight, Rect(0,0,dwWidth-1,dwHeight-1));
          if r <> S_OK then begin // the call failed!
            DeleteDC(tempDC);
            DeleteObject(tempBitmap);
          end
          else begin
            // we now have the bitmap blitted, let's get it off the dc and clean up.
            SelectObject(tempDC, OldBitmap);
            DeleteDC(tempDC);
            SaveBitmapToStream(tempBitmap,OutStream);
            DeleteObject(tempBitmap);
            Result := True;
          end
        end
      end
    end
  end
end;

procedure ConvertSPV(var instream, outstream: TStream);
var
  Buf: array[8..$6B] of byte;
  i:integer;
begin
  instream.Read(Buf,SizeOf(Buf));
  for i := Low(Buf) to High(Buf) do
    Buf[i] := not Buf[i];
  outstream.Write(Buf,SizeOf(Buf));
end;

procedure ConvertWBD(var instream,outstream: TStream);
var
  Buf: array[8..$6B+$64] of byte;
  i:integer;
  oldPos:integer;
begin
  oldPos := instream.Position;
  instream.Read(Buf,SizeOf(Buf));
  for i := 8 to $6B do
    Buf[i] := (not Buf[i]) xor (Buf[i+$64]) xor $A4;
  outstream.Write(Buf,SizeOf(Buf)-$64);
  instream.Seek(oldPos+$6C-8,soFromBeginning);
end;

procedure ConvertWB1(var instream,outstream: TStream);
var
  Buf: array[8..$6B+$64] of byte;
  i:integer;
  oldPos:integer;
begin
  oldPos := instream.Position;
  instream.Read(Buf,SizeOf(Buf));
  for i := 8 to $6B do
    Buf[i] := (not Buf[i]) xor (Buf[i+$64]) xor $F2;
  outstream.Write(Buf,SizeOf(Buf)-$64);
  instream.Seek(oldPos+$6C-8, soFromBeginning);
end;

function ConversionNeeded(var instream: TStream): Boolean;
//checks if header conversion is required. Resets stream position to where it was before calling
var
  ID: array[1..8] of AnsiChar;
  OrigPos: Int64;
begin
  OrigPos := instream.Position;
  instream.Read(ID,SizeOf(ID));
  Result := (ID = idSPV) or (ID = idWBD) or (ID = idWB1);
  instream.Seek(OrigPos, soFromBeginning);
end;

procedure ConvertHeader(var instream,outstream: TStream);
//decodes header and moves on (doesn't reset stream position)
var
  ID: array[1..8] of AnsiChar;
begin
  instream.Read(ID,SizeOf(ID));
// decode header according to ID
  if ID = idSPV then
    ConvertSPV(instream,outstream)
  else
  if ID = idWBD then
    ConvertWBD(instream,outstream)
  else
  if ID = idWB1 then
    ConvertWB1(instream,outstream)
  else //probably jpeg
    outstream.Write(ID,SizeOf(ID))
end;

procedure TWebShots.EncodeHeader(var instream,outstream: TStream);
//encodes header according to FFileType and moves on (doesn't reset stream position)
begin
// encode header according to ID
  case FFileType of
    wsftSPV: ConvertSPV(instream,outstream);
    wsftWBD: ConvertWBD(instream,outstream);
    wsftWB1: ConvertWB1(instream,outstream);
//  else <Raise exception?>
  end
end;

procedure TWebShots.CreateEmpty;
begin
  if FFileType = wsftUnknown then
    raise Exception.Create('File type not specified');
(*$IFNDEF WBZ_OUTPUT*)
  if FFileType = wsftWBZ then
    raise Exception.Create('WBZ file creation is not supported');
(*$ENDIF*)

  FStream := TFileStream.Create(FFileName,fmCreate or fmShareDenyWrite); //This may result in an exception, which should be handled by caller

  try
    if FFileType = wsftWBC then begin //write WBC header
      FillChar(FWBCHeader,SizeOf(FWBCHeader),#0);
      FWBCHeader.ID := idWBC;
      FWBCHeader.FirstPic := SizeOf(FWBCHeader);
      FWBCHeader._unknown := 1;
      FStream.Seek(0, soFromBeginning);
      if FStream.Write(FWBCHeader, SizeOf(FWBCHeader)) < SizeOf(FWBCHeader) then
        raise Exception.Create('Error writing to file');
      FID := FWBCHeader.ID;
      FCount := FWBCHeader.FileCount;
    end
  (*$IFDEF WBZ_OUTPUT*)
    else if FFileType = wsftWBZ then begin //write WBZ header
      FillChar(FWBZHeader,SizeOf(FWBZHeader),#0);
      FWBZHeader.ID := idWBZ;
      FStream.Seek(0, soFromBeginning);
      if FStream.Write(FWBZHeader, SizeOf(FWBZHeader)) < SizeOf(FWBZHeader) then
        raise Exception.Create('Error writing to file');
      FCount := 0;
    end
  (*$ENDIF*)
    else begin
      if FFileType = wsftSPV then FIDStr := idSPV else
      if FFileType = wsftWBD then FIDStr := idWBD else
      if FFileType = wsftWB1 then FIDStr := idWB1 else
        raise Exception.Create('Unknown file type');
      if FStream.Write(FIDStr,SizeOf(FIDStr)) < SizeOf(FIDStr) then
        raise Exception.Create('Error writing to file');
      FCount := 0;
    end
  finally
    FreeAndNil(FStream);
  end;
end;

constructor TWebShots.Create(const FileName: TFileName; FileType: Cardinal = wsftUnknown);
var
  Header: array[0..7] of AnsiChar;
begin
  inherited Create;

  FFileName  := FileName;
  FID        := 0;
  FIDStr     := '';
  FFileType  := FileType;
  FCount     := 0;

  //open file
  if not FileExists(FileName) then //Try to create a new, empty one
    CreateEmpty;
  FStream := TFileStream.Create(FileName,fmOpenRead or fmShareDenyNone);

  //get ID
  if FStream.Read(Header, SizeOf(Header)) < SizeOf(Header) then
    raise Exception.Create('Unexpected end of file');

  Move(Header,FID,SizeOf(FID));
  Move(Header,FIDStr,SizeOf(FIDStr));

  //set FileType according to ID
  if FID = idWBC then FFileType := wsftWBC else
  if FID = idWBZ then FFileType := wsftWBZ else
  if FIDStr = idSPV then FFileType := wsftSPV else
  if FIDStr = idWBD then FFileType := wsftWBD else
  if FIDStr = idWB1 then FFileType := wsftWB1 else
  FFileType := wsftUnknown;

  if FFileType = wsftWBC then begin //buffer WBC header
    FStream.Seek(0, soFromBeginning);
    if FStream.Read(FWBCHeader, SizeOf(FWBCHeader)) < SizeOf(FWBCHeader) then
      raise Exception.Create('Unexpected end of file');
    FCount := FWBCHeader.FileCount;
  end
  else if FFileType = wsftWBZ then begin //buffer WBZ header
    FStream.Seek(0, soFromBeginning);
    if FStream.Read(FWBZHeader, SizeOf(FWBZHeader)) < SizeOf(FWBZHeader) then
      raise Exception.Create('Unexpected end of file');
    FCount := 1;
  end
  else begin //Stand-alone file
    if FStream.Position = FStream.Size then //empty file created by CreateEmpty
      FCount := 0
    else
      FCount := 1
  end
end;

destructor TWebShots.Destroy;
begin
  FStream.Free;
  inherited;
end;

function TWebShots.GetFileType:Cardinal;
begin
  Result := FFileType;
end;

procedure TWebShots.SetFileType(FileType: Cardinal);
begin
  if FStream = nil then
    FFileType := FileType
  else
    //raise exception?
  ;
end;

function TWebShots.GetHandle: THandle;
begin
  if (FStream <> nil) and (FStream is THandleStream) then
    Result := THandleStream(FStream).Handle
  else
    Result := 0
end;

function TWebShots.GetEntry(PicIndex: Integer; var Entry: TWSEntry): Boolean;
var
  WBCPicHeader: TWBCPicHeader;
  WBZPicHeader: TWBZPicHeader;
  JPGName: AnsiString;
begin
  Result := False;

  FillChar(Entry,SizeOf(Entry),#0);

  if FFileType = wsftWBC then begin
    if picIndex > FWBCHeader.FileCount then Exit;
    if picIndex < Low(FWBCHeader.DirEntry) then Exit;

    with FWBCHeader.DirEntry[picIndex] do begin
      FStream.Seek(Start, soFromBeginning);
      FStream.Read(WBCPicHeader, SizeOf(WBCPicHeader));
    end;

    Entry.Date := UnixToDateTime(FWBCHeader.DirEntry[picIndex].Date);
    Move(WBCPicHeader.FileName,Entry.FileName,SizeOf(Entry.FileName));
    Move(WBCPicHeader.Title,Entry.Title,SizeOf(Entry.Title));
    Entry.DataSize := WBCPicHeader.PicSize;

    Result := True;
  end
  else if FFileType = wsftWBZ then begin //we don't yet know how to handle these...
    if picIndex = 1 then begin
      Entry.Date := FileGetDate(THandleStream(FStream).Handle);

      FStream.Seek(FWBZHeader.FirstPic, soFromBeginning);
      FStream.Read(WBZPicHeader, SizeOf(WBZPicHeader));

      Move(WBZPicHeader.FileName,Entry.FileName,SizeOf(Entry.FileName));
      Entry.DataSize := WBZPicHeader.DataSize;

      Result := True;
    end
  end
  else begin //stand-alone file
    if (FCount = 1) and (picIndex = 1) then begin //ok, go on
      Entry.Date := FileGetDate(THandleStream(FStream).Handle);

      JPGName := ChangeFileExt(ExtractFileName(FFileName),'.jpg');
      if Length(JPGName) <= SizeOf(Entry.FileName) then
        Move(JPGName[1], Entry.FileName, Length(JPGName))
      else
        Move(JPGName[1], Entry.FileName, SizeOf(Entry.FileName));
      Entry.DataSize := FStream.Size-8;

      Result := True;
    end
  end
end;


function TWebShots.Extract(PicIndex: Integer; FileName: TFileName): Boolean;
var
  WBCPicHeader: TWBCPicHeader;
  WBZPicHeader: TWBZPicHeader;
  OutStream: TStream;
begin
  Result := False;

  if FFileType = wsftWBC then begin
    if picIndex > FWBCHeader.FileCount then Exit;
    if picIndex < Low(FWBCHeader.DirEntry) then Exit;

    with FWBCHeader.DirEntry[picIndex] do begin
      FStream.Seek(Start, soFromBeginning);
      FStream.Read(WBCPicHeader, SizeOf(WBCPicHeader));
      FStream.Seek(Start + WBCPicHeader.HeaderSize, soFromBeginning);
    end;

    OutStream := TFileStream.Create(FileName, fmCreate);
    try
      OutStream.CopyFrom(FStream,WBCPicHeader.PicSize);
      Result := True;
    finally
      OutStream.Free;
    end;
  end
  else if FFileType = wsftWBZ then begin //we don't yet know how to handle these...
    if picIndex = 1 then begin //ok, go on
      FStream.Seek(FWBZHeader.FirstPic,soFromBeginning);
      FStream.Read(WBZPicHeader, SizeOf(WBZPicHeader));
      FStream.Seek(FWBZHeader.FirstPic + WBZPicHeader.HeaderSize,soFromBeginning);

      OutStream := TFileStream.Create(FileName, fmCreate);
      try
        OutStream.CopyFrom(FStream,WBZPicHeader.DataSize);
        Result := True;
      finally
        OutStream.Free;
      end;
    end
  end
  else begin //stand-alone file
    if (FCount = 1) and (picIndex = 1) then begin //ok, go on
      OutStream := TFileStream.Create(FileName, fmCreate);
      FStream.Seek(0,soFromBeginning);
      try
        // decode header
        ConvertHeader(FStream,OutStream);
        OutStream.CopyFrom(FStream,FStream.Size-FStream.Position);
        Result := True;
      finally
        OutStream.Free;
      end
    end
  end
end;

function TWebShots.Add(FileName: TFileName): Boolean;
var
  WBCPicHeader: TWBCPicHeader;
//  WBZPicHeader: TWBZPicHeader;
  InStream: TStream;
  WBCPicHeaderPos, WBCPicDataPos, WBCPicThumbPos: Int64;
  Name: AnsiString;
  MemStream: TStream;
begin
  Result := False;
  Name := ExtractFileName(FileName);

  if FFileType = wsftWBC then begin
    if FWBCHeader.FileCount < 160 then begin // There is a free slot, go on
      //re-open WBC stream in write mode
      FStream.Free;
      try
        FStream := TFileStream.Create(FFileName, fmOpenReadWrite or fmShareDenyNone);

        //open input stream
        InStream := TFileStream.Create(FileName, fmOpenRead or fmShareDenyNone);
        try
          FillChar(WBCPicHeader,SizeOf(WBCPicHeader),#0);
          WBCPicHeader.ID := idWBCPic;
          WBCPicHeader.HeaderSize := SizeOf(WBCPicHeader);
          WBCPicHeader.PicSize := InStream.Size;
          WBCPicHeader.Size := WBCPicHeader.HeaderSize + WBCPicHeader.PicSize;
          if Length(Name) <= SizeOf(WBCPicHeader.FileName) then
            Move(Name[1], WBCPicHeader.FileName, Length(Name))
          else
            Move(Name[1], WBCPicHeader.FileName, SizeOf(WBCPicHeader.FileName));
          if Length(Name) <= SizeOf(WBCPicHeader.Title) then
            Move(Name[1], WBCPicHeader.Title, Length(Name))
          else
            Move(Name[1], WBCPicHeader.Title, SizeOf(WBCPicHeader.Title));
          Move(ExtractFileExt(Name)[2],WBCPicHeader.Kind,Length(ExtractFileExt(Name))-1);
//          WBCPicHeader.ThumbnailSize := 11078;
          WBCPicHeader.DateAdded := DateTimeToUnix(Now);

          inc(FWBCHeader.FileCount);
          FCount := FWBCHeader.FileCount;

          FillChar(FWBCHeader.DirEntry[FWBCHeader.FileCount],SizeOf(FWBCHeader.DirEntry[FWBCHeader.FileCount]),#0);
          FWBCHeader.DirEntry[FWBCHeader.FileCount].Start := FStream.Size;

          FStream.Seek(0, soFromEnd);
          WBCPicHeaderPos := FStream.Position;
          FStream.Write(WBCPicHeader,SizeOf(WBCPicHeader));
          WBCPicDataPos := FStream.Position;
          FStream.CopyFrom(InStream, InStream.Size);
          WBCPicThumbPos := FStream.Position;

          //Create thumbnail
          InStream.Seek(0,soFromBeginning);

          if ConversionNeeded(InStream) then begin
            MemStream := TMemoryStream.Create;
            try
              ConvertHeader(InStream,MemStream); // decode header
              MemStream.CopyFrom(InStream,InStream.Size-InStream.Position);
(*$IFDEF DEBUG*)     TMemoryStream(MemStream).SaveToFile('memstream.dmp'); (*$ENDIF*)
              MemStream.Seek(0,soFromBeginning);
              if SaveThumbnailToStream(MemStream,FStream) then
                WBCPicHeader.ThumbnailSize := FStream.Position - WBCPicThumbPos
              else
                WBCPicHeader.ThumbnailSize := 0;
            finally
              MemStream.Free;
            end;
          end
          else begin
            if SaveThumbnailToStream(InStream,FStream) then
              WBCPicHeader.ThumbnailSize := FStream.Position - WBCPicThumbPos
            else
              WBCPicHeader.ThumbnailSize := 0;
          end;

          FWBCHeader.DirEntry[FWBCHeader.FileCount].Length := InStream.Size + SizeOf(WBCPicHeader) + WBCPicHeader.ThumbnailSize;
          FWBCHeader.DirEntry[FWBCHeader.FileCount].Date := WBCPicHeader.DateAdded;

          FStream.Seek(0, soFromBeginning);
          FStream.Write(FWBCHeader,Sizeof(FWBCHeader));

          FStream.Seek(WBCPicHeaderPos, soFromBeginning);
          FStream.Write(WBCPicHeader,SizeOf(WBCPicHeader));

          Result := True;
        finally
          InStream.Free;
        end;
      finally
        //re-open WBC stream in read mode
        FStream.Free;
        FStream := TFileStream.Create(FFileName, fmOpenRead or fmShareDenyNone);
      end
    end
  end
  else if FFileType = wsftWBZ then begin
    //TODO: We don't yet know how to add files to WBZ
  end
  else begin //stand-alone file
    if FCount = 0 then begin //We don't want to overwrite picture data
      //re-open WBC stream in write mode
      FStream.Free;
      try
        FStream := TFileStream.Create(FFileName, fmOpenReadWrite or fmShareDenyNone);
        FStream.Seek(8,soFromBeginning); //skip 8-byte ID

        //open input stream
        InStream := TFileStream.Create(FileName, fmOpenRead or fmShareDenyNone);
        try
          //Encode header, copy rest of file, delete incomplete file on error
          EncodeHeader(InStream,FStream);
          FStream.CopyFrom(InStream,InStream.Size-InStream.Position);
          Result := True;
        finally
          InStream.Free;
        end;
      finally
        //re-open WBC stream in read mode
        FStream.Free;
        FStream := TFileStream.Create(FFileName, fmOpenRead or fmShareDenyNone);
      end
    end
  end
end;

end.
